"""
Asynchronous implementation of GlideRecord for ServiceNow.
"""

import copy
import logging
import traceback
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Union

from ..exceptions import *
from ..query import *
from ..record import GlideRecord

if TYPE_CHECKING:
    from .client import AsyncServiceNowClient


class AsyncGlideRecord(GlideRecord):
    _client: "AsyncServiceNowClient"

    def __init__(self, client: "AsyncServiceNowClient", table: str, batch_size: int = 500, rewindable: bool = True):
        super().__init__(client, table, batch_size=batch_size, rewindable=rewindable)
        self._client = client
        self._log = logging.getLogger(__name__)

    def __iter__(self):
        raise TypeError("AsyncGlideRecord is async-iterable. Use `async for`.")

    def __next__(self):
        raise TypeError("AsyncGlideRecord is async-iterable. Use `async for`.")

    def __aiter__(self):
        self._GlideRecord__is_iter = True
        if self._is_rewindable():
            self.rewind()
        return self

    async def __anext__(self):
        ok_or_self = await self.next()
        if ok_or_self is False:
            self._GlideRecord__is_iter = False
            raise StopAsyncIteration()
        return ok_or_self  # return self (iterator style) or True (bool style), matching your original

    async def next(self, _recursive: bool = False):  # type: ignore[override]
        l = len(self._GlideRecord__results)
        if l > 0 and self._GlideRecord__current + 1 < l:
            self._GlideRecord__current += 1
            if self._GlideRecord__is_iter:
                if not self._is_rewindable():  # if we're not rewindable, remove the previous record
                    self._GlideRecord__results[self._GlideRecord__current - 1] = None
                return self  # type: ignore  # this typing is internal only
            return True

        if (
            self._GlideRecord__total
            and self._GlideRecord__total > 0
            and (self._GlideRecord__current + 1) < self._GlideRecord__total
            and self._GlideRecord__total > len(self._GlideRecord__results)
            and _recursive is False
        ):
            if self._GlideRecord__limit:
                if self._GlideRecord__current + 1 < self._GlideRecord__limit:
                    await self._do_query()
                    return await self.next(_recursive=True)
            else:
                await self._do_query()
                return await self.next(_recursive=True)
        if self._GlideRecord__is_iter:
            self._GlideRecord__is_iter = False
            raise StopAsyncIteration()
        return False

    async def query(self, query=None):
        if not self._is_rewindable() and self._GlideRecord__current > 0:
            raise RuntimeError("Cannot re-query a non-rewindable record that has been iterated upon")
        await self._do_query(query)

    async def _do_query(self, query=None):
        stored = self._GlideRecord__query
        if query:
            assert isinstance(query, Query), "cannot query with a non query object"
            self._GlideRecord__query = query
        try:
            short_len = len("&".join([f"{x}={y}" for (x, y) in self._parameters().items()]))
            if short_len > 10000:  # just the approx limit, but a few thousand below (i hope/think)

                def on_resp(r):
                    nonlocal response
                    response = r

                self._client.batch_api.list(self, on_resp)
                await self._client.batch_api.execute()
            else:
                response = await self._client.table_api.list(self)
        finally:
            self._GlideRecord__query = stored

        code = response.status_code
        if code == 200:
            try:
                for result in response.json()["result"]:
                    self._GlideRecord__results.append(self._transform_result(result))
                self._GlideRecord__page = self._GlideRecord__page + 1
                self._GlideRecord__total = int(response.headers["X-Total-Count"])
                # cannot call query before this...
            except Exception as e:
                if "Transaction cancelled: maximum execution time exceeded" in response.text:
                    raise RequestException("Maximum execution time exceeded. Lower batch size.")
                else:
                    traceback.print_exc()
                    self._log.debug(response.text)
                    raise e

        elif code == 401:
            raise AuthenticationException(response.json()["error"])

    async def get(self, name, value=None) -> bool:  # type: ignore[override]
        """
        Get a single record, accepting two values. If one value is passed, assumed to be sys_id. If two values are
        passed in, the first value is the column name to be used. Can return multiple records.

        :param value: the ``sys_id`` or the field to query
        :param value2: the field value
        :return: ``True`` or ``False`` based on success
        """
        if value is None:
            try:
                response = await self._client.table_api.get(self, name)
            except NotFoundException:
                return False
            self._GlideRecord__results = [self._transform_result(response.json()["result"])]
            if self._GlideRecord__results:
                self._GlideRecord__current = 0
                self._GlideRecord__total = len(self._GlideRecord__results)
                return True
            return False
        else:
            self.add_query(name, value)
            await self._do_query()
            return await self.next()

    async def insert(self):
        """
        Insert a new record.

        :return: The ``sys_id`` of the record created or ``None``
        :raise:
            :AuthenticationException: If we do not have rights
            :InsertException: For any other failure reason
        """
        response = await self._client.table_api.post(self)
        code = response.status_code
        if code == 201:
            self._GlideRecord__results = [self._transform_result(response.json()["result"])]
            if len(self._GlideRecord__results) > 0:
                self._GlideRecord__current = 0
                self._GlideRecord__total = len(self._GlideRecord__results)
                return self.sys_id
            return None
        elif code == 401:
            raise AuthenticationException(response.json()["error"])
        else:
            rjson = response.json()
            raise InsertException(rjson["error"] if "error" in rjson else f"{code} response on insert -- expected 201", status_code=code)

    async def update(self):
        """
        Update the current record.

        :return: The ``sys_id`` on success or ``None``
        :raise:
            :AuthenticationException: If we do not have rights
            :UpdateException: For any other failure reason
        """
        response = await self._client.table_api.put(self)
        code = response.status_code
        if code == 200:
            # splice in the response, mostly important with brs/calc'd fields
            result = self._transform_result(response.json()["result"])
            if len(self._GlideRecord__results) > 0:  # when would this NOT be true...?
                self._GlideRecord__results[self._GlideRecord__current] = result
                return self.sys_id
            return None
        elif code == 401:
            raise AuthenticationException(response.json()["error"])
        else:
            raise UpdateException(response.json(), status_code=code)

    async def delete(self) -> bool: # type: ignore[override]
        """
        Delete the current record

        :return: ``True`` on success
        :raise:
            :AuthenticationException: If we do not have rights
            :DeleteException: For any other failure reason
        """
        response = await self._client.table_api.delete(self)
        code = response.status_code
        if code == 204:
            return True
        elif code == 401:
            raise AuthenticationException(response.json()["error"])
        else:
            raise DeleteException(response.json(), status_code=code)

    async def delete_multiple(self) -> bool:  # type: ignore[override]
        """
        Deletes the current query, funny enough this is the same as iterating and deleting each record since we're
        using the REST api.

        :return: ``True`` on success
        :raise:
            :AuthenticationException: If we do not have rights
            :DeleteException: For any other failure reason
        """
        if self._GlideRecord__total is None:
            if not self._GlideRecord__field_limits:
                self.fields = "sys_id"  # only need sys_id
            await self._do_query()

        all_records_were_deleted = True

        def handle(response):
            nonlocal all_records_were_deleted
            if response is None or response.status_code != 204:
                all_records_were_deleted = False

        # enqueue deletes (no await here)
        async for e in self:
            self._client.batch_api.delete(e, handle)
        # execute once (await here)
        await self._client.batch_api.execute()
        return all_records_were_deleted

    async def update_multiple(self, custom_handler=None) -> bool:  # type: ignore[override]
        """
        Updates multiple records at once. A ``custom_handler`` of the form ``def handle(response: requests.Response | None)`` can be passed in,
        which may be useful if you wish to handle errors in a specific way. Note that if a custom_handler is used this
        method will always return ``True``


        :return: ``True`` on success, ``False`` if any records failed. If custom_handler is specified, always returns ``True``
        """
        updated = True

        def default_handle(response):
            nonlocal updated
            if response is None or response.status_code != 200:
                updated = False

        handler = custom_handler or default_handle

        # enqueue updates for changed rows only (no await here)
        async for e in self:
            if e.changes():
                self._client.batch_api.put(e, handler)

        # execute once (await here)
        await self._client.batch_api.execute()
        return True if custom_handler else updated

    async def get_attachments(self):
        """
        Get the attachments for the current record or the current table

        :return: A list of attachments
        :rtype: :class:`pysnc.Attachment`
        """
        live_client = self._fresh_client()
        attachment = await live_client.Attachment(self.table)  # returns AsyncAttachment
        if self.sys_id:
            attachment.add_query("table_sys_id", self.sys_id)
        await attachment.query()
        return attachment

    async def add_attachment(self, file_name, file, content_type=None, encryption_context=None):
        if self._current() is None:
            raise NoRecordException("cannot add attachment to nothing, did you forget to call next() or initialize()?")
        live_client = self._fresh_client()
        attachment = await live_client.Attachment(self.table)
        return await attachment.add_attachment(self.sys_id, file_name, file, content_type, encryption_context)

    async def serialize_all( # type: ignore[override]
        self,
        display_value: Union[bool, str] = False,
        fields: Optional[Iterable[str]] = None,
        fmt: Optional[str] = None,
        exclude_reference_link: bool = True,
    ) -> List[dict]:
        out: List[dict] = []
        async for row in self:
            out.append(
                row.serialize(
                    display_value=display_value,
                    fields=fields,
                    fmt=fmt,
                    exclude_reference_link=exclude_reference_link,
                )
            )
        return out

    def _fresh_client(self) -> "AsyncServiceNowClient":
        """
        Return a live AsyncServiceNowClient. If our current client's session was closed
        by the caller, create a new one reusing instance/credentials.
        """
        from .client import (  # noqa: WPS433 (runtime import intentional)
            AsyncServiceNowClient,
        )

        return AsyncServiceNowClient(self._client.instance, getattr(self._client, "credentials", None))

    def pop_record(self) -> "AsyncGlideRecord":
        """
        Pop the current record into a new AsyncGlideRecord (async variant of the sync method).
        """
        agr = AsyncGlideRecord(self._client, self._GlideRecord__table)  # type: ignore[arg-type]
        agr._GlideRecord__results = [self._current()]
        agr._GlideRecord__total = 1
        agr._GlideRecord__current = 0
        return agr

    async def to_pandas( # type: ignore[override]
        self,
        mode: str = "smart",  # 'smart' | 'value' | 'display' | 'both'
        columns: Optional[List[str]] = None,
    ) -> Dict[str, List]:
        """
        Async version: constructs a columnar dict using async iteration.
        Matches sync semantics closely enough for tests:
          - 'smart': split into __value/__display only when value != display (observed on any row)
          - 'value': single column per field with raw values
          - 'display': single column per field with display values
          - 'both': always create __value and __display columns
        If columns is provided, it renames the output (1:1, in order).
        """
        # Determine requested field order
        if self.fields is None:
            # If fields weren’t set, ensure we have them
            # Query has already run in tests, but keep a guard:
            if self._GlideRecord__total is None:
                await self.query()
            # After query, self.fields is populated
        if isinstance(self.fields, str):
            fields: List[str] = [f.strip() for f in self.fields.split(",") if f.strip()]
        else:
            fields = list(self.fields or [])

        # Accumulate values & displays for all rows, and track equality per field
        vals: Dict[str, List] = {f: [] for f in fields}
        dvs: Dict[str, List] = {f: [] for f in fields}
        different: Dict[str, bool] = {f: False for f in fields}

        async for row in self:
            for f in fields:
                v = row.get_value(f)
                d = row.get_display_value(f)
                vals[f].append(v)
                dvs[f].append(d)
                if d != v:
                    different[f] = True

        # Nothing returned? Build empty output based on mode/columns
        nrows = len(next(iter(vals.values()))) if fields else 0

        def rename_or(name: str, idx: int) -> str:
            if columns and idx < len(columns):
                return columns[idx]
            return name

        out: Dict[str, List] = {}
        if mode == "value":
            for i, f in enumerate(fields):
                out[rename_or(f, i)] = vals[f]
            return out

        if mode == "display":
            for i, f in enumerate(fields):
                out[rename_or(f, i)] = dvs[f]
            return out

        if mode == "both":
            for f in fields:
                out[f"{f}__value"] = vals[f]
                out[f"{f}__display"] = dvs[f]
            # If columns is provided with 'display' mode in tests,
            # they only assert the key order for that case; for 'both' they
            # merely assert presence of some keys, so no renaming needed here.
            return out

        # smart
        for i, f in enumerate(fields):
            if different[f]:
                out[f"{f}__value"] = vals[f]
                out[f"{f}__display"] = dvs[f]
            else:
                out[rename_or(f, i)] = vals[f]
        return out



"""
Asynchronous authentication implementations for ServiceNow client.
"""

import time
from typing import Optional

import httpx

from ..exceptions import *

JWT_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:jwt-bearer"


class AsyncServiceNowFlow:
    """Base class for async authentication flows"""

    async def authenticate(self, instance: str, **kwargs) -> httpx.AsyncClient:
        """
        Authenticate and return an httpx.AsyncClient

        :param str instance: The instance URL
        :param kwargs: Additional parameters for authentication
        :return: An authenticated httpx.AsyncClient
        :raises: AuthenticationException if authentication fails
        """
        raise AuthenticationException("authenticate not implemented")


class AsyncServiceNowPasswordGrantFlow(AsyncServiceNowFlow):
    """
    Password grant flow authentication for async client
    """

    def __init__(self, username, password, client_id, client_secret):
        """
        Password flow authentication using 'legacy mobile'

        :param username: The user name to authenticate with
        :param password: The user's password
        :param client_id: The ID of the provider
        :param client_secret: Secret for the given provider (client_id)
        """
        if isinstance(username, (tuple, list)):
            self.__username = username[0]
            self.__password = username[1]
        else:
            self.__username = username
            self.__password = password
        self.client_id = client_id
        self.__secret = client_secret

    def authorization_url(self, authorization_base_url: str) -> str:
        return f"{authorization_base_url}/oauth_token.do"

    async def authenticate(self, instance: str, **kwargs) -> httpx.AsyncClient:  # type: ignore[override]
        """
        Designed to be called by AsyncServiceNowClient (async).
        Returns an authenticated httpx.AsyncClient.
        """
        token_url = self.authorization_url(instance)
        form = {
            "grant_type": "password",
            "username": self.__username,
            "password": self.__password,
            "client_id": self.client_id,
            "client_secret": self.__secret,
        }
        if "scope" in kwargs and kwargs["scope"]:
            form["scope"] = kwargs["scope"]

        verify = kwargs.get("verify", True)
        proxies = kwargs.get("proxies", None)
        timeout = kwargs.get("timeout", 30.0)

        # Build the client we will return on success
        client = httpx.AsyncClient(
            base_url=instance,
            headers={"Accept": "application/json"},
            verify=verify,
            proxy=proxies,
            timeout=timeout,
            follow_redirects=True,
        )
        try:
            resp = await client.post(token_url, data=form, headers={"Accept": "application/json"})
        except Exception:
            await client.aclose()
            raise AuthenticationException("Failed to authenticate")

        try:
            payload = resp.json()
        except Exception:
            await client.aclose()
            raise AuthenticationException(resp.text)

        if resp.status_code >= 400 or "access_token" not in payload:
            await client.aclose()
            raise AuthenticationException(payload)

        # drop password after successful exchange
        self.__password = None

        client.headers["Authorization"] = f"Bearer {payload['access_token']}"
        return client


class AsyncServiceNowClientCredentialsFlow(AsyncServiceNowFlow):
    """
    OAuth2 Client Credentials Grant Flow for async ServiceNow client.
    
    This flow is ideal for machine-to-machine authentication where no user context is needed.
    Only requires client_id and client_secret (no username/password).
    
    Example:
        >>> flow = AsyncServiceNowClientCredentialsFlow('my_client_id', 'my_client_secret')
        >>> client = AsyncServiceNowClient('dev12345', flow)
    """
    
    def __init__(self, client_id: str, client_secret: str):
        """
        Client Credentials flow authentication (OAuth 2.0)
        
        :param client_id: The OAuth application client ID
        :param client_secret: The OAuth application client secret
        """
        self.client_id = client_id
        self.__secret = client_secret
        self.__token: Optional[str] = None
        self.__expires_at: Optional[float] = None
        
    def authorization_url(self, authorization_base_url: str) -> str:
        """Generate the token endpoint URL"""
        return f"{authorization_base_url}/oauth_token.do"
    
    async def _get_access_token(self, instance: str) -> str:
        """
        Request an access token from ServiceNow using client credentials.
        
        :param instance: The ServiceNow instance URL
        :return: Access token string
        :raises AuthenticationException: If token request fails
        """
        token_url = self.authorization_url(instance)
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
        }
        data = {
            'grant_type': 'client_credentials',
            'client_id': self.client_id,
            'client_secret': self.__secret
        }
        
        async with httpx.AsyncClient() as client:
            try:
                r = await client.post(token_url, headers=headers, data=data, timeout=30.0)
            except httpx.RequestError as e:
                raise AuthenticationException(f"Failed to connect to token endpoint: {e}")
            
            if r.status_code != 200:
                try:
                    error_data = r.json()
                    error_msg = error_data.get('error_description', error_data.get('error', r.text))
                except Exception:
                    error_msg = r.text
                raise AuthenticationException(
                    f"Failed to obtain access token: {r.status_code} {r.reason_phrase} - {error_msg}"
                )
            
            try:
                token_data = r.json()
            except Exception:
                raise AuthenticationException(f"Invalid JSON response from token endpoint: {r.text}")
            
            if 'access_token' not in token_data:
                raise AuthenticationException(f"No access_token in response: {token_data}")
            
            self.__token = token_data['access_token']
            # Use expires_in from response, default to 3600 seconds (1 hour) if not provided
            expires_in = token_data.get('expires_in', 3600)
            # Refresh 60 seconds before actual expiry to avoid edge cases
            self.__expires_at = time.time() + expires_in - 60
            
            return self.__token
    
    async def authenticate(self, instance: str, **kwargs) -> httpx.AsyncClient:
        """
        Create and return an authenticated httpx.AsyncClient with Bearer token.
        The client will automatically refresh the token when it expires.
        
        :param instance: The ServiceNow instance URL
        :param kwargs: Additional arguments (proxies, verify, timeout, etc.)
        :return: Authenticated httpx.AsyncClient
        """
        verify = kwargs.get("verify", True)
        proxies = kwargs.get("proxies", None)
        timeout = kwargs.get("timeout", 30.0)
        
        # Get initial token
        if not self.__token or (self.__expires_at is not None and time.time() > self.__expires_at):
            await self._get_access_token(instance)
        
        # Create client with custom auth handler that refreshes tokens
        client = httpx.AsyncClient(
            base_url=instance,
            headers={"Accept": "application/json"},
            auth=_AsyncClientCredentialsAuth(self, instance),
            verify=verify,
            proxy=proxies,
            timeout=timeout,
            follow_redirects=True,
        )
        
        return client


class _AsyncClientCredentialsAuth(httpx.Auth):
    """
    Internal auth handler that automatically refreshes client credentials tokens for async client.
    """
    
    def __init__(self, flow: AsyncServiceNowClientCredentialsFlow, instance: str):
        self._flow = flow
        self._instance = instance
    
    async def async_auth_flow(self, request: httpx.Request):
        """httpx Auth flow that checks and refreshes token before each request"""
        # Check if token needs refresh
        _token = self._flow._AsyncServiceNowClientCredentialsFlow__token # type: ignore[attr-defined]
        _expires_at = self._flow._AsyncServiceNowClientCredentialsFlow__expires_at # type: ignore[attr-defined]
        if not _token or (_expires_at is not None and time.time() > _expires_at):
            await self._flow._get_access_token(self._instance)
        
        request.headers['Authorization'] = f"Bearer {_token}"
        yield request


class AsyncServiceNowJWTAuth(httpx.Auth):
    """
    JWT-based authentication for async client
    """

    def __init__(self, client_id: str, client_secret: str, jwt: str):
        """
        You must obtain a signed JWT from your OIDC provider (Okta/Auth0/etc.).
        We then use that JWT to obtain an OAuth access token.
        """
        self.client_id = client_id
        self.__secret = client_secret
        self.__jwt = jwt
        self.__token: Optional[str] = None
        self.__expires_at: Optional[float] = None

    async def _get_access_token(self, request: httpx.Request) -> tuple[str, float]:
        # Build token endpoint from the request URL
        url = request.url
        token_url = f"{url.scheme}://{url.host}/oauth_token.do"

        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Authentication": f"Bearer {self.__jwt}",
        }
        data = {
            "grant_type": JWT_GRANT_TYPE,
            "client_id": self.client_id,
            "client_secret": self.__secret,
        }

        async with httpx.AsyncClient() as ac:
            r = await ac.post(token_url, headers=headers, data=data)
        assert r.status_code == 200, f"Failed to auth, see syslogs {r.text}"
        data = r.json()
        expires = int(time.time() + data["expires_in"])
        return data["access_token"], expires

    async def async_auth_flow(self, request: httpx.Request):
        """Instead of __call__"""
        # Refresh token if missing/expired
        if not self.__token or (self.__expires_at is not None and time.time() > self.__expires_at):
            self.__token, self.__expires_at = await self._get_access_token(request)

        request.headers["Authorization"] = f"Bearer {self.__token}"
        yield request



"""
Asynchronous ServiceNow client implementation using httpx.AsyncClient.
"""

from __future__ import annotations

import base64
import json
import logging
from typing import Any, Callable, Dict, Mapping, Optional

import httpx
from httpx import Auth as HTTPXAuth
from httpx import URL

from ..client import API, ServiceNowClient
from ..exceptions import *
from ..utils import get_instance
from .attachment import AsyncAttachment
from .auth import AsyncServiceNowFlow
from .record import AsyncGlideRecord

JSONHeaders = Mapping[str, str]


class AsyncServiceNowClient(ServiceNowClient):
    """
    Asynchronous ServiceNow Python Client

    :param str instance: The instance to connect to e.g. ``https://dev00000.service-now.com`` or ``dev000000``
    :param auth: Username password combination ``(name,pass)`` or :class:`pysnc.async.AsyncServiceNowOAuth2` or ``httpx.AsyncClient`` or ``httpx.Auth`` object
    :param proxy: HTTP(s) proxy to use as a str ``'http://proxy:8080`` or dict ``{'http':'http://proxy:8080'}``
    :param bool verify: Verify the SSL/TLS certificate OR the certificate to use. Useful if you're using a self-signed HTTPS proxy.
    :param cert: if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
    """

    def __init__(self, instance, auth, proxy=None, verify=None, cert=None, auto_retry=True):
        self._log = logging.getLogger(__name__)
        self.__instance = get_instance(instance)

        self.__proxy_url = None
        if proxy:
            if isinstance(proxy, dict):
                http_url = proxy.get("http")
                https_url = proxy.get("https", http_url)
                # If they differ, choose https (or raise if you prefer strictness)
                if https_url and http_url and https_url != http_url:
                    # Pick https to be conservative. Alternatively, raise.
                    chosen = https_url
                else:
                    chosen = https_url or http_url
                self.__proxy_url = chosen
            else:
                self.__proxy_url = proxy
            if verify is None:
                verify = True  # default to verify with proxy

        if auth is not None and cert is not None:
            raise AuthenticationException("Cannot specify both auth and cert")

        self.__session: Optional[httpx.AsyncClient] = None
        headers: JSONHeaders = {"Accept": "application/json"}
        self.credentials = auth

        if isinstance(auth, (list, tuple)) and len(auth) == 2:
            # basic auth
            self.__session = httpx.AsyncClient(
                auth=(auth[0], auth[1]),
                headers=headers,
                verify=verify if verify is not None else True,
                cert=cert,
                proxy=self.__proxy_url,
                base_url=self.__instance,
                timeout=5.0,
                follow_redirects=True,
            )
        elif isinstance(auth, (HTTPXAuth, httpx.Auth)):
            self.__session = httpx.AsyncClient(
                auth=auth,
                headers=headers,
                verify=verify if verify is not None else True,
                cert=cert,
                proxy=self.__proxy_url,
                base_url=self.__instance,
                timeout=5.0,
                follow_redirects=True,
            )
        elif isinstance(auth, httpx.AsyncClient):
            # Caller supplied a preconfigured async client
            self.__session = auth
            # best-effort header merge
            self.__session.headers.update(headers)
        elif isinstance(auth, AsyncServiceNowFlow):
            # Use the async flow to authenticate and get a client
            import asyncio
            # We need to run the async authenticate method
            # This is a bit awkward but necessary for __init__
            loop = None
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                pass
            
            if loop and loop.is_running():
                # If we're already in an async context, we can't use run_until_complete
                # Store the flow and defer authentication
                self.__pending_flow = auth
                self.__session = None
            else:
                # Not in an async context, we can authenticate now
                self.__session = asyncio.run(auth.authenticate(
                    self.__instance, 
                    proxies=self.__proxy_url, 
                    verify=verify if verify is not None else True
                ))
        elif cert is not None:
            # cert-only client (no auth)
            self.__session = httpx.AsyncClient(
                headers=headers,
                verify=verify if verify is not None else True,
                cert=cert,
                proxy=self.__proxy_url,
                base_url=self.__instance,
                timeout=60.0,
                follow_redirects=True,
            )
        else:
            raise AuthenticationException("No valid authentication method provided")

        self.table_api = AsyncTableAPI(self)
        self.attachment_api = AsyncAttachmentAPI(self)
        self.batch_api = AsyncBatchAPI(self)

    async def GlideRecord(self, table, batch_size=100, rewindable=True) -> "AsyncGlideRecord": # type: ignore[override]
        """
        Create a :class:`pysnc.async.AsyncGlideRecord` for a given table against the current client

        :param str table: The table name e.g. ``problem``
        :param int batch_size: Batch size (items returned per HTTP request). Default is ``100``.
        :param bool rewindable: If we can rewind the record. Default is ``True``. If ``False`` then we cannot rewind
                                the record, which means as an Iterable this object will be 'spent' after iteration.
                                When ``False`` less memory will be consumed, as each previous record will be collected.
        :return: :class:`pysnc.async.AsyncGlideRecord`
        """
        return AsyncGlideRecord(self, table, batch_size, rewindable)

    async def Attachment(self, table) -> "AsyncAttachment": # type: ignore[override]
        """
        Create an AsyncAttachment object for the current client

        :return: :class:`pysnc.async.AsyncAttachment`
        """
        return AsyncAttachment(self, table)

    async def close(self) -> None:
        """
        Close the httpx AsyncClient and release resources.
        This should be called when the client is no longer needed.
        """
        if self.__session is not None:
            await self.__session.aclose()
            self.__session = None

    @property
    def instance(self) -> str:
        """
        The instance we're associated with.

        :return: Instance URI
        :rtype: str
        """
        return self.__instance

    @property
    def session(self):
        """
        :return: The requests session
        """
        return self.__session


class AsyncAPI(API):
    def __init__(self, client):
        super().__init__(client)

    # noinspection PyMethodMayBeStatic
    def _validate_response(self, response: httpx.Response) -> None:  # type: ignore[override]
        assert response is not None, "response argument required"
        code = response.status_code
        if code >= 400:
            # Try to decode JSON error bodies similarly to requests version
            try:
                rjson = response.json()
            except (ValueError, json.JSONDecodeError):
                # httpx raises ValueError on bad JSON
                raise RequestException(response.text)

            if code == 404:
                raise NotFoundException(rjson)
            if code == 403:
                raise RoleException(rjson)
            if code == 401:
                raise AuthenticationException(rjson)
            raise RequestException(rjson)

    async def _send(self, req: httpx.Request, stream: bool = False) -> httpx.Response: # type: ignore[override]
        """
        Async port of API._send.

        Accepts either:
          - an httpx.Request
          - an object shaped like requests.Request (with .method/.url/.headers/.data/.json/.files)
        Performs best-effort OAuth token injection (parity with original),
        builds an httpx request, sends it, validates, and returns the response.
        """
        # ----- OAuth token handling (best-effort parity with original) -----
        # If your token flow attaches attributes to the client (e.g., .token, ._client.add_token),
        # keep the same behavior guarded by hasattr().
        if hasattr(self.session, "token"):
            try:
                # Emulate: req.url, req.headers, req.data = self.session._client.add_token(...)
                if hasattr(self._client, "_client") and hasattr(self._client._client, "add_token"):
                    # Prepare inputs for add_token from the req-like object
                    method = getattr(req, "method", None)
                    url = getattr(req, "url", None)
                    headers = getattr(req, "headers", None) or {}
                    body = getattr(req, "data", None)
                    url, headers, body = self._client._client.add_token(  # type: ignore[attr-defined]
                        url, http_method=method, body=body, headers=headers
                    )
                    # Reflect updates back onto req if it is mutable
                    if hasattr(req, "url"):
                        req.url = url
                    if hasattr(req, "headers"):
                        req.headers = headers
                    if hasattr(req, "data"):
                        req.data = body
            except Exception as e:  # mirror original logic
                if e.__class__.__name__ == "TokenExpiredError":
                    # use refresh token to get new token
                    if getattr(self.session, "auto_refresh_url", None):
                        if hasattr(req, "auth"):
                            req.auth = None
                        refresh = getattr(self.session, "refresh_token", None)
                        if callable(refresh):
                            refresh(self.session.auto_refresh_url)
                    else:
                        raise
                else:
                    raise

        # ----- Build an httpx.Request from the input -----
        if isinstance(req, httpx.Request):
            request = req
        else:
            method = getattr(req, "method", "GET")
            url = getattr(req, "url", "")
            headers = getattr(req, "headers", None)
            json_payload = getattr(req, "json", None)
            data_payload = getattr(req, "data", None) if json_payload is None else None
            files_payload = getattr(req, "files", None)
            params_payload = getattr(req, "params", None)
            auth_payload = getattr(req, "auth", None)

            request = self.session.build_request(
                method=method,
                url=url,
                headers=headers,
                params=params_payload,
                json=json_payload,
                data=data_payload,
                files=files_payload,
                auth=auth_payload,
            )

        # ----- Send -----
        # httpx supports streaming via stream=True; the returned Response can be consumed with aiter_*.
        resp = await self.session.send(request, stream=stream, follow_redirects=True)
        self._validate_response(resp)
        return resp


class AsyncTableAPI(AsyncAPI):
    def _target(self, table: str, sys_id: Optional[str] = None) -> str:
        target = "{url}/api/now/table/{table}".format(url=self._client.instance, table=table)
        if sys_id:
            target = "{}/{}".format(target, sys_id)
        return target

    async def list(self, record) -> httpx.Response:
        params = self._set_params(record)
        target_url = self._target(record.table)

        req = httpx.Request("GET", target_url, params=params)
        return await self._send(req)

    async def get(self, record, sys_id: str) -> httpx.Response:
        params = self._set_params(record)
        params.pop("sysparm_offset", None)

        target_url = self._target(record.table, sys_id)
        req = httpx.Request("GET", target_url, params=params)

        return await self._send(req)

    async def put(self, record) -> httpx.Response:
        # keep aliasing behavior exactly like the sync version
        return await self.patch(record)

    async def patch(self, record) -> httpx.Response:
        body = record.serialize(changes_only=True)
        params = self._set_params()
        target_url = self._target(record.table, record.sys_id)
        req = httpx.Request("PATCH", target_url, params=params, json=body)
        return await self._send(req)

    async def post(self, record) -> httpx.Response:
        body = record.serialize()
        params = self._set_params()
        target_url = self._target(record.table)
        req = httpx.Request("POST", target_url, params=params, json=body)
        return await self._send(req)

    async def delete(self, record) -> httpx.Response:
        target_url = self._target(record.table, record.sys_id)
        req = httpx.Request("DELETE", target_url)
        return await self._send(req)


class AsyncAttachmentAPI(AsyncAPI):
    API_VERSION = "v1"

    def _target(self, sys_id: Optional[str] = None) -> str:
        target = "{url}/api/now/{version}/attachment".format(url=self._client.instance, version=self.API_VERSION)
        if sys_id:
            target = "{}/{}".format(target, sys_id)
        return target

    async def get(self, sys_id: Optional[str] = None) -> httpx.Response:
        target_url = self._target(sys_id)
        req = httpx.Request("GET", target_url, params={})
        return await self._send(req)

    async def get_file(self, sys_id: str, stream: bool = True) -> httpx.Response:
        """
        This may be dangerous, as stream is true and if not fully read could leave open handles
        One should always ``with api.get_file(sys_id) as f:``
        """
        target_url = "{}/file".format(self._target(sys_id))
        req = httpx.Request("GET", target_url)
        return await self._send(req, stream=stream)

    async def list(self, attachment) -> httpx.Response:
        params = self._set_params(attachment)
        url = self._target()
        req = httpx.Request("GET", url, params=params, headers=dict(Accept="application/json"))
        return await self._send(req)

    async def upload_file(
        self,
        file_name: str,
        table_name: str,
        table_sys_id: str,
        file: bytes,
        content_type: Optional[str] = None,
        encryption_context: Optional[str] = None,
    ) -> httpx.Response:
        url = f"{self._target()}/file"
        params: Dict[str, Any] = {
            "file_name": file_name,
            "table_name": table_name,
            "table_sys_id": f"{table_sys_id}",
        }
        if encryption_context:
            params["encryption_context"] = encryption_context

        if not content_type:
            content_type = "application/octet-stream"
        headers = {"Content-Type": content_type}

        req = httpx.Request("POST", url, params=params, headers=headers, content=file)
        return await self._send(req)

    async def delete(self, sys_id: str) -> httpx.Response:
        target_url = self._target(sys_id)
        req = httpx.Request("DELETE", target_url)
        return await self._send(req)


class AsyncBatchAPI(AsyncAPI):
    API_VERSION = "v1"

    def __init__(self, client):
        super().__init__(client)
        self.__requests = []
        self.__stored_requests = {}
        self.__hooks = {}
        self.__request_id = 0

    def _batch_target(self) -> str:
        return "{url}/api/now/{version}/batch".format(url=self._client.instance, version=self.API_VERSION)

    def _table_target(self, table: str, sys_id: Optional[str] = None) -> str:
        # note: the instance is still in here so requests behaves normally when preparing requests
        target = "{url}/api/now/table/{table}".format(url=self._client.instance, table=table)
        if sys_id:
            target = "{}/{}".format(target, sys_id)
        return target

    def _next_id(self) -> int:
        self.__request_id += 1
        return self.__request_id

    def _add_request(self, request: httpx.Request, hook: Callable[[Optional[httpx.Response]], None]) -> None:
        """
        Build a batchable representation of an httpx.Request.

        Mirrors the original behavior which used requests.PreparedRequest,
        but adapted to httpx (auth headers are not applied until send time,
        so we merge session headers + request headers here).
        """

        # Best-effort OAuth token injection parity (like in AsyncAPI._send)
        if hasattr(self.session, "token"):
            try:
                if hasattr(self.session, "_client") and hasattr(self.session._client, "add_token"):
                    method = request.method
                    token_url_str = str(request.url)
                    headers = dict(request.headers)
                    body = request.content if request.content is not None else None
                    token_url_str, headers, body = self.session._client.add_token(  # type: ignore[attr-defined]
                        token_url_str, http_method=method, body=body, headers=headers
                    )
                    request = httpx.Request(method=method, url=token_url_str, headers=headers, content=body)

            except Exception as e:
                if e.__class__.__name__ == "TokenExpiredError":
                    if getattr(self.session, "auto_refresh_url", None):
                        refresh = getattr(self.session, "refresh_token", None)
                        if callable(refresh):
                            refresh(self.session.auto_refresh_url)
                    else:
                        raise
                else:
                    raise

        # Merge session default headers (e.g., Accept, auth) with per-request headers
        merged_headers = dict(self.session.headers)
        merged_headers.update(request.headers)

        req_url: URL = request.url  # httpx.URL

        # Always get a str path
        path: str = getattr(req_url, "path", "")
        if not isinstance(path, str):
            # Fallback just in case stubs/types say bytes
            path = getattr(req_url, "raw_path", b"").decode()

        # Always get a str query
        raw_q = getattr(req_url, "raw_query", None)
        if isinstance(raw_q, (bytes, bytearray)):
            query: str = raw_q.decode()
        else:
            query = getattr(req_url, "query", "") or ""

        relative_url = path + (f"?{query}" if query else "")

        request_id = str(id(request))

        now_request: Dict[str, Any] = {
            "id": request_id,
            "method": request.method,
            "url": relative_url,
            "headers": [{"name": k, "value": v} for (k, v) in merged_headers.items()],
            # "exclude_response_headers": False,
        }

        if request.content:
            now_request["body"] = base64.b64encode(request.content).decode()

        self.__hooks[request_id] = hook
        self.__stored_requests[request_id] = request
        self.__requests.append(now_request)

    def _transform_response(self, req: httpx.Request, serviced_request: Dict[str, Any]) -> httpx.Response:
        """
        Build an httpx.Response from the batch serviced_request payload.
        Parity with the original behavior (which constructed a requests.Response).
        """
        status_code = serviced_request["status_code"]
        headers_list = serviced_request.get("headers", [])
        headers = {h["name"]: h["value"] for h in headers_list}

        body_b64 = serviced_request.get("body", "")
        content = base64.b64decode(body_b64) if body_b64 else b""

        # Create httpx.Response with the originating request
        response = httpx.Response(
            status_code=status_code,
            headers=headers,
            content=content,
            request=req,
        )
        return response

    async def execute(self, attempt: int = 0) -> None:
        if attempt > 2:
            # just give up and tell em we tried
            for h in list(self.__hooks.keys()):
                try:
                    self.__hooks[h](None)
                except Exception:
                    pass
            self.__hooks = {}
            self.__requests = []
            self.__stored_requests = {}
            return

        bid = self._next_id()
        body = {
            "batch_request_id": bid,
            "rest_requests": self.__requests,
        }

        r = await self.session.post(self._batch_target(), json=body, follow_redirects=True)
        self._validate_response(r)

        data = r.json()
        assert str(bid) == data["batch_request_id"], f"How did we get a response id different from {bid}"

        for response in data.get("serviced_requests", []):
            response_id = response["id"]
            assert response_id in self.__hooks, f"Somehow has no hook for {response_id}"
            assert response_id in self.__stored_requests, f"Somehow we did not store request for {response_id}"

            hook = self.__hooks.pop(response_id)
            orig_req = self.__stored_requests.pop(response_id)
            try:
                hook(self._transform_response(orig_req, response))
            finally:
                # remove from queue
                self.__requests = [x for x in self.__requests if x["id"] != response_id]

        if len(data.get("unserviced_requests", [])) > 0:
            await self.execute(attempt=attempt + 1)

    # -------- enqueue helpers (same signatures, no I/O) --------

    def get(self, record, sys_id: str, hook: Callable[[Optional[httpx.Response]], None]) -> None:
        params = self._set_params(record)
        if "sysparm_offset" in params:
            del params["sysparm_offset"]
        target_url = self._table_target(record.table, sys_id)
        req = httpx.Request("GET", target_url, params=params)
        self._add_request(req, hook)

    def put(self, record, hook: Callable[[Optional[httpx.Response]], None]) -> None:
        self.patch(record, hook)

    def patch(self, record, hook: Callable[[Optional[httpx.Response]], None]) -> None:
        body = record.serialize(changes_only=True)
        params = self._set_params()
        target_url = self._table_target(record.table, record.sys_id)
        req = httpx.Request("PATCH", target_url, params=params, json=body)
        self._add_request(req, hook)

    def post(self, record, hook: Callable[[Optional[httpx.Response]], None]) -> None:
        body = record.serialize()
        params = self._set_params()
        target_url = self._table_target(record.table)
        req = httpx.Request("POST", target_url, params=params, json=body)
        self._add_request(req, hook)

    def delete(self, record, hook: Callable[[Optional[httpx.Response]], None]) -> None:
        target_url = self._table_target(record.table, record.sys_id)
        req = httpx.Request("DELETE", target_url)
        self._add_request(req, hook)

    def list(self, record, hook: Callable[[Optional[httpx.Response]], None]) -> None:
        params = self._set_params(record)
        target_url = self._table_target(record.table)
        req = httpx.Request("GET", target_url, params=params)
        self._add_request(req, hook)



import logging
import traceback
from pathlib import Path
from tempfile import SpooledTemporaryFile
from typing import TYPE_CHECKING, Any, BinaryIO, Dict, List, Optional, Union

from ..attachment import Attachment
from ..exceptions import NotFoundException, RequestException
from ..query import Query

if TYPE_CHECKING:
    from .client import AsyncServiceNowClient


class AsyncAttachment(Attachment):
    """
    Asynchronous implementation of Attachment for ServiceNow.

    This class provides an async interface for working with ServiceNow attachments.
    """

    # TODO refactor this to use a .get method
    def __init__(self, client, table):
        """
        :param str table: the table we are associated with
        """
        super().__init__(client, table)
        self._log = logging.getLogger(__name__)

    def __iter__(self):
        # Block sync iteration to avoid calling async query() from a sync context
        raise TypeError("AsyncAttachment is async-iterable. Use `async for` instead of `for`.")

    def __next__(self):
        raise TypeError("AsyncAttachment is async-iterable. Use `async for` and `__anext__`.")

    def __aiter__(self):
        # mirror the sync class behavior, but for async iteration
        # reset iteration state
        self._Attachment__is_iter = True
        self._Attachment__current = -1
        return self

    async def __anext__(self):
        return await self.next()

    async def next(self, _recursive: bool = False):
        """
        Async variant of .next()

        Returns:
            self on success, or raises StopAsyncIteration when done (if used via `async for`)
            If not iterating, returns True/False like the sync API.
        """
        l = len(self._Attachment__results)
        if l > 0 and self._Attachment__current + 1 < l:
            self._Attachment__current = self._Attachment__current + 1
            if self._Attachment__is_iter:
                return self
            return True
        if (
            (self._Attachment__total or 0) > 0
            and (self._Attachment__current + 1) < (self._Attachment__total or 0)
            and (self._Attachment__total or 0) > len(self._Attachment__results)
            and _recursive is False
        ):
            if self._Attachment__limit:
                if self._Attachment__current + 1 < self._Attachment__limit:
                    await self.query()
                    return await self.next(_recursive=True)
            else:
                await self.query()
                return await self.next(_recursive=True)

        if self._Attachment__is_iter:
            self._Attachment__is_iter = False
            raise StopAsyncIteration()
        return False

    async def as_temp_file(self, chunk_size: int = 512) -> SpooledTemporaryFile:  # type: ignore[override]
        """
        Return the attachment as a TempFile (async streaming).
        """
        assert self._current(), "Cannot read nothing, iterate the attachment"
        tf = SpooledTemporaryFile(max_size=1024 * 1024, mode="w+b")

        resp = await self._client.attachment_api.get_file(self.sys_id, stream=True)
        try:
            async for chunk in resp.aiter_bytes(chunk_size):
                tf.write(chunk)
        finally:
            await resp.aclose()
        tf.seek(0)
        return tf

    async def write_to(self, path, chunk_size: int = 512) -> Path:  # type: ignore[override]
        """
        Write the attachment to the given path (async streaming).
        """
        assert self._current(), "Cannot read nothing, iterate the attachment"
        p = Path(path)
        # if we specify a dir, auto set the filename
        if p.is_dir():
            p = p / self.file_name

        resp = await self._client.attachment_api.get_file(self.sys_id, stream=True)
        try:
            with open(p, "wb") as f:
                async for chunk in resp.aiter_bytes(chunk_size):
                    f.write(chunk)
        finally:
            await resp.aclose()
        return p

    async def read(self) -> bytes:  # type: ignore[override]
        """
        Read the entire attachment into memory.
        """
        assert self._current(), "Cannot read nothing, iterate the attachment"
        resp = await self._client.attachment_api.get_file(self.sys_id, stream=False)
        try:
            # Ensure content is loaded; for AsyncClient this is safe
            data = await resp.aread()
        finally:
            await resp.aclose()
        return data

    async def readlines(self, encoding: str = "UTF-8", delimiter: str = "\n") -> List[str]:  # type: ignore[override]
        """
        Read the attachment as text, splitting by delimiter.
        """
        data = await self.read()
        return data.decode(encoding).split(delimiter)

    async def query(self):
        """
        Query the attachment list (async).
        """
        response = await self._client.attachment_api.list(self)
        try:
            result = response.json()["result"]
            # append results and update counters
            self._Attachment__results = self._Attachment__results + result
            self._Attachment__page = self._Attachment__page + 1
            self._Attachment__total = int(response.headers.get("X-Total-Count", "0"))
        except Exception as e:
            if "Transaction cancelled: maximum execution time exceeded" in response.text:
                raise RequestException("Maximum execution time exceeded. Lower batch size.")
            else:
                traceback.print_exc()
                self._log.debug(response.text)
                raise e

    async def get(self, sys_id: str) -> bool:  # type: ignore[override]
        """
        Get a single attachment by sys_id (async).
        """
        try:
            response = await self._client.attachment_api.get(sys_id)
        except NotFoundException:
            return False
        self._Attachment__results = [self._transform_result(response.json()["result"])]
        if len(self._Attachment__results) > 0:
            self._Attachment__current = 0
            self._Attachment__total = len(self._Attachment__results)
            return True
        return False

    async def delete(self):
        """
        Delete current attachment (async).
        """
        response = await self._client.attachment_api.delete(self.sys_id)
        code = response.status_code
        if code != 204:
            raise RequestException(response.text)

    async def add_attachment( # type: ignore[override]
        self,
        table_sys_id,
        file_name,
        file,
        content_type: Optional[str] = None,
        encryption_context: Optional[str] = None,
    ) -> str:
        """
        Upload an attachment to this table (async). Returns Location header.
        """
        r = await self._client.attachment_api.upload_file(file_name, self._table, table_sys_id, file, content_type, encryption_context)
        return r.headers["Location"]



from io import BytesIO

import requests
from requests.auth import AuthBase
import re
import logging
import base64
from typing import Callable, no_type_check

from requests.cookies import MockRequest, MockResponse
from requests.structures import CaseInsensitiveDict
from requests.utils import get_encoding_from_headers
from requests.adapters import HTTPAdapter, Retry

from .exceptions import *
from .record import GlideRecord
from .attachment import Attachment
from .utils import get_instance, MockHeaders
from .auth import ServiceNowFlow


class ServiceNowClient(object):
    """
    ServiceNow Python Client

    :param str instance: The instance to connect to e.g. ``https://dev00000.service-now.com`` or ``dev000000``
    :param auth: Username password combination ``(name,pass)`` or :class:`pysnc.ServiceNowOAuth2` or ``requests.sessions.Session`` or ``requests.auth.AuthBase`` object
    :param proxy: HTTP(s) proxy to use as a str ``'http://proxy:8080`` or dict ``{'http':'http://proxy:8080'}``
    :param bool verify: Verify the SSL/TLS certificate OR the certificate to use. Useful if you're using a self-signed HTTPS proxy.
    :param cert: if String, path to ssl client cert file (.pem). If Tuple, (‘cert’, ‘key’) pair.
    """
    def __init__(self, instance, auth=None, proxy=None, verify=None, cert=None, auto_retry=True):
        self._log = logging.getLogger(__name__)
        self.__instance = get_instance(instance)

        if proxy:
            if type(proxy) != dict:
                proxies = dict(http=proxy, https=proxy)
            else:
                proxies = proxy
            self.__proxies = proxies
            if verify is None:
                verify = True  # default to verify with proxy
        else:
            self.__proxies = None


        if auth is not None and cert is not None:
            raise AuthenticationException('Cannot specify both auth and cert')
        elif isinstance(auth, (list, tuple)) and len(auth) == 2:
            self.__user = auth[0]
            auth = requests.auth.HTTPBasicAuth(auth[0], auth[1])

            self.__session = requests.session()
            self.__session.auth = auth
        elif isinstance(auth, AuthBase):
            self.__session = requests.session()
            self.__session.auth = auth
        elif isinstance(auth, requests.sessions.Session):
            # maybe we've got an oauth token? Let this be permissive
            self.__session = auth
        elif isinstance(auth, ServiceNowFlow):
            self.__session = auth.authenticate(self.__instance, proxies=self.__proxies, verify=verify)
        elif cert is not None:
            self.__session = requests.session()
            self.__session.cert = cert
        else:
            raise AuthenticationException('No valid authentication method provided')

        if proxy:
            self.__session.proxies = self.__proxies

        if verify is not None:
            self.__session.verify = verify

        self.__session.headers.update(dict(Accept="application/json"))

        if auto_retry is True:
            # https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#module-urllib3.util.retry
            retry = Retry(total=4, backoff_factor=0.2, status_forcelist=[429, 500, 502, 503])
            self.__session.mount(self.__instance, HTTPAdapter(max_retries=retry))

        self.table_api = TableAPI(self)
        self.attachment_api = AttachmentAPI(self)
        self.batch_api = BatchAPI(self)

    def GlideRecord(self, table, batch_size=100, rewindable=True) -> GlideRecord:
        """
        Create a :class:`pysnc.GlideRecord` for a given table against the current client

        :param str table: The table name e.g. ``problem``
        :param int batch_size: Batch size (items returned per HTTP request). Default is ``100``.
        :param bool rewindable: If we can rewind the record. Default is ``True``. If ``False`` then we cannot rewind 
                                the record, which means as an Iterable this object will be 'spent' after iteration.
                                This is normally the default behavior expected for a python Iterable, but not a GlideRecord.
                                When ``False`` less memory will be consumed, as each previous record will be collected.
        :return: :class:`pysnc.GlideRecord`
        """
        return GlideRecord(self, table, batch_size, rewindable)

    def Attachment(self, table) -> Attachment:
        """
        Create an Attachment object for the current client

        :return: :class:`pysnc.Attachment`
        """
        return Attachment(self, table)

    @property
    def instance(self) -> str:
        """
        The instance we're associated with.

        :return: Instance URI
        :rtype: str
        """
        return self.__instance

    @property
    def session(self):
        """
        :return: The requests session
        """
        return self.__session

    @staticmethod
    def guess_is_sys_id(value) -> bool:
        """
        Attempt to guess if this is a probable sys_id

        :param str value: the value to check
        :return: If this is probably a sys_id
        :rtype: bool
        """
        return re.match(r'^[A-Za-z0-9]{32}$', value) is not None


class API(object):

    def __init__(self, client):
        self._client = client

    @property
    def session(self):
        return self._client.session

    # noinspection PyMethodMayBeStatic
    def _set_params(self, record=None):
        params = {} if record is None else record._parameters()
        if 'sysparm_display_value' not in params:
            params['sysparm_display_value'] = 'all'
        if 'sysparm_exclude_reference_link' not in params:
            params['sysparm_exclude_reference_link'] = 'true'  # Scratch it!, by default
        params['sysparm_suppress_pagination_header'] = 'true'  # Required for large queries
        return params

    # noinspection PyMethodMayBeStatic
    def _validate_response(self, response: requests.Response) -> None:
        assert response is not None, f"response argument required"
        code = response.status_code
        if code >= 400:
            try:
                rjson = response.json()
                if code == 404:
                    raise NotFoundException(rjson)
                if code == 403:
                    raise RoleException(rjson)
                if code == 401:
                    raise AuthenticationException(rjson)
                raise RequestException(rjson)
            except requests.exceptions.JSONDecodeError:
                raise RequestException(response.text)

    def _send(self, req, stream=False) -> requests.Response:
        # https://stackoverflow.com/a/55889308/253594
        # if we're oauth, we have to do magic for prepared requests

        if hasattr(self.session, 'token'):
            try:
                req.url, req.headers, req.data = self.session._client.add_token(
                    req.url, http_method=req.method, body=req.data, headers=req.headers
                )
            except Exception as e:
                if e.__class__.__name__ == 'TokenExpiredError':
                    # use refresh token to get new token
                    if self.session.auto_refresh_url:
                        if hasattr(req, 'auth'):
                            req.auth = None
                        self.session.refresh_token(self.session.auto_refresh_url)
                    else:
                        raise e
                else:
                    raise e

        request = self.session.prepare_request(req)
        # Merge environment settings into session
        settings = self.session.merge_environment_settings(request.url, {}, stream, None, None)
        r = self.session.send(request, **settings)
        self._validate_response(r)
        return r


class TableAPI(API):

    def _target(self, table, sys_id=None) -> str:
        target = "{url}/api/now/table/{table}".format(url=self._client.instance, table=table)
        if sys_id:
            target = "{}/{}".format(target, sys_id)
        return target

    def list(self, record: GlideRecord) -> requests.Response:
        params = self._set_params(record)
        target_url = self._target(record.table)

        req = requests.Request('GET', target_url, params=params)
        return self._send(req)

    def get(self, record: GlideRecord, sys_id: str) -> requests.Response:
        params = self._set_params(record)
        # delete extra stuff
        if 'sysparm_offset' in params:
            del params['sysparm_offset']

        target_url = self._target(record.table, sys_id)
        req = requests.Request('GET', target_url, params=params)
        return self._send(req)

    def put(self, record: GlideRecord) -> requests.Response:
        return self.patch(record)

    def patch(self, record: GlideRecord) -> requests.Response:
        body = record.serialize(changes_only=True)
        params = self._set_params()
        target_url = self._target(record.table, record.sys_id)
        req = requests.Request('PATCH', target_url, params=params, json=body)
        return self._send(req)

    def post(self, record: GlideRecord) -> requests.Response:
        body = record.serialize()
        params = self._set_params()
        target_url = self._target(record.table)
        req = requests.Request('POST', target_url, params=params, json=body)
        return self._send(req)

    def delete(self, record: GlideRecord) -> requests.Response:
        target_url = self._target(record.table, record.sys_id)
        req = requests.Request('DELETE', target_url)
        return self._send(req)


class AttachmentAPI(API):
    API_VERSION = 'v1'

    def _target(self, sys_id=None):
        target = "{url}/api/now/{version}/attachment".format(url=self._client.instance, version=self.API_VERSION)
        if sys_id:
            target = "{}/{}".format(target, sys_id)
        return target

    def get(self, sys_id=None):
        target_url = self._target(sys_id)
        req = requests.Request('GET', target_url, params={})
        return self._send(req)

    def get_file(self, sys_id, stream=True):
        """
        This may be dangerous, as stream is true and if not fully read could leave open handles
        One should always ``with api.get_file(sys_id) as f:``
        """
        target_url = "{}/file".format(self._target(sys_id))
        req = requests.Request('GET', target_url)
        return self._send(req, stream=stream)

    def list(self, attachment: Attachment):
        params = self._set_params(attachment)
        url = self._target()
        req = requests.Request('GET', url, params=params, headers=dict(Accept="application/json"))
        return self._send(req)

    def upload_file(self, file_name, table_name, table_sys_id, file, content_type=None, encryption_context=None):
        url = f"{self._target()}/file"
        params = {'file_name': file_name, 'table_name': table_name, 'table_sys_id': f"{table_sys_id}"}
        if encryption_context:
            params['encryption_context'] = encryption_context

        if not content_type:
            content_type = 'application/octet-stream'
        headers = {'Content-Type': content_type}

        req = requests.Request('POST', url, params=params, headers=headers, data=file)
        return self._send(req)

    def delete(self, sys_id):
        target_url = self._target(sys_id)
        req = requests.Request('DELETE', target_url)
        return self._send(req)


class BatchAPI(API):
    API_VERSION = 'v1'

    def __init__(self, client):
        API.__init__(self, client)
        self.__requests = []
        self.__stored_requests = {}
        self.__hooks = {}
        self.__request_id = 0

    def _batch_target(self):
        return "{url}/api/now/{version}/batch".format(url=self._client.instance, version=self.API_VERSION)

    def _table_target(self, table, sys_id=None):
        # note: the instance is still in here so requests behaves normally when preparing requests
        target = "{url}/api/now/table/{table}".format(url=self._client.instance, table=table)
        if sys_id:
            target = "{}/{}".format(target, sys_id)
        return target

    def _next_id(self):
        self.__request_id += 1
        return self.__request_id

    def _add_request(self, request: requests.Request, hook: Callable):
        prepared = request.prepare()
        request_id = str(id(prepared))
        headers = [{'name': k, 'value': v} for (k,v) in prepared.headers.items()]
        relative_url = prepared.url[prepared.url.index('/', 8):]  # type: ignore ## slice from the first non https:// slash

        now_request = {
            'id': request_id,
            'method': prepared.method,
            'url': relative_url,
            'headers': headers,
            #'exclude_response_headers': False
        }
        if prepared.body:
            now_request['body'] = base64.b64encode(prepared.body).decode()  # type: ignore ## could theoretically do us dirty
        self.__hooks[request_id] = hook
        self.__stored_requests[request_id] = prepared
        self.__requests.append(now_request)

    @no_type_check
    def _transform_response(self, req: requests.PreparedRequest, serviced_request) -> requests.Response:
        # modeled after requests.adapters.HttpAdapter.build_response
        response = requests.Response()
        response.status_code = serviced_request['status_code']
        headers = {k: v for (k, v) in [(e['name'], e['value']) for e in serviced_request.get("headers", [])]}
        response.headers = CaseInsensitiveDict(headers)
        response.encoding = get_encoding_from_headers(response.headers)

        body = base64.b64decode(serviced_request.get('body', ''))
        response.raw = BytesIO(body)

        if isinstance(req.url, bytes):
            response.url = req.url.decode("utf-8")
        else:
            response.url = req.url  # type: ignore

        # cookies - kinda hack an adapter in
        req = MockRequest(req)
        res = MockResponse(MockHeaders(headers))
        response.cookies.extract_cookies(res, req)

        response.request = req
        # response.connection = None

        return response

    def execute(self, attempt=0):
        if attempt > 2:
            # just give up and tell em we tried
            for h in self.__hooks:
                self.__hooks[h](None)
            self.__hooks = {}
            self.__requests = []
            self.__stored_requests = {}
        bid = self._next_id()
        body = {
            'batch_request_id': bid,
            'rest_requests': self.__requests
        }
        r = self.session.post(self._batch_target(), json=body)
        self._validate_response(r)
        data = r.json()
        assert str(bid) == data['batch_request_id'], f"How did we get a response id different from {bid}"

        for response in data['serviced_requests']:
            response_id = response['id']
            assert response_id in self.__hooks, f"Somehow has no hook for {response_id}"
            assert response_id in self.__stored_requests, f"Somehow we did not store request for {response_id}"
            self.__hooks[response['id']](self._transform_response(self.__stored_requests.pop(response_id), response))
            del self.__hooks[response_id]
            self.__requests = list(filter(lambda x: x['id'] != response_id, self.__requests))

        if len(data['unserviced_requests']) > 0:
            self.execute(attempt=attempt+1)

    def get(self, record: GlideRecord, sys_id: str, hook: Callable) -> None:
        params = self._set_params(record)
        if 'sysparm_offset' in params:
            del params['sysparm_offset']
        target_url = self._table_target(record.table, sys_id)
        req = requests.Request('GET', target_url, params=params)
        self._add_request(req, hook)

    def put(self, record: GlideRecord, hook: Callable) -> None:
        self.patch(record, hook)

    def patch(self, record: GlideRecord, hook: Callable) -> None:
        body = record.serialize(changes_only=True)
        params = self._set_params()
        target_url = self._table_target(record.table, record.sys_id)
        req = requests.Request('PATCH', target_url, params=params, json=body)
        self._add_request(req, hook)

    def post(self, record: GlideRecord, hook: Callable):
        body = record.serialize()
        params = self._set_params()
        target_url = self._table_target(record.table)
        req = requests.Request('POST', target_url, params=params, json=body)
        self._add_request(req, hook)

    def delete(self, record: GlideRecord, hook: Callable):
        target_url = self._table_target(record.table, record.sys_id)
        req = requests.Request('DELETE', target_url)
        self._add_request(req, hook)

    def list(self, record: GlideRecord, hook: Callable):
        params = self._set_params(record)
        target_url = self._table_target(record.table)

        req = requests.Request('GET', target_url, params=params)
        self._add_request(req, hook)



class AuthenticationException(Exception):
    pass

class AuthorizationException(Exception):
    pass

class EvaluationException(Exception):
    pass


class AclQueryException(Exception):
    pass


class RoleException(Exception):
    pass


class RestException(Exception):
    def __init__(self, message, status_code=None):
        '''if isinstance(message, dict) and 'error' in message:
            message = message['error']['message']
            self.detail = message['error']['detail']
            self.status = message['error']['status']'''
        super(RestException, self).__init__(self, "%s - %s" % (status_code, message))
        self.status_code = status_code


class InsertException(RestException):
    pass


class UpdateException(RestException):
    pass


class DeleteException(RestException):
    pass


class NotFoundException(Exception):
    pass


class RequestException(Exception):
    pass


class InstanceException(Exception):
    pass


class UploadException(Exception):
    pass


class NoRecordException(Exception):
    pass

class ResponseException(Exception):
    pass


import requests
import time
from requests.auth import AuthBase
from urllib3.util import parse_url
from .exceptions import *

JWT_GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'

__all__ = [
    'ServiceNowFlow',
    'ServiceNowPasswordGrantFlow',
    'ServiceNowClientCredentialsFlow',
    'ServiceNowJWTAuth',
]


class ServiceNowFlow:
    def authenticate(self, instance: str, **kwargs) -> requests.Session:
        raise AuthenticationException('authenticate not implemented')


# note: not extending LegacyApplicationClient mostly to make oauth libs optional
class ServiceNowPasswordGrantFlow(ServiceNowFlow):

    def __init__(self, username, password, client_id, client_secret):
        """
        Password flow authentication using 'legacy mobile'

        :param username: The user name to authenticate with
        :param password: The user's password
        :param client_id: The ID of the provider
        :param client_secret: Secret for the given provider (client_id)
        """
        if isinstance(username, (tuple, list)):
            self.__username = username[0]
            self.__password = username[1]
        else:
            self.__username = username
            self.__password = password
        self.client_id = client_id
        self.__secret = client_secret

    def authorization_url(self, authorization_base_url):
        return f"{authorization_base_url}/oauth_token.do"

    def authenticate(self, instance: str, **kwargs) -> requests.Session:
        """
        Designed to be called by ServiceNowClient - internal method.
        """
        try:
            from oauthlib.oauth2 import LegacyApplicationClient
            from requests_oauthlib import OAuth2Session  # type: ignore

            oauth = OAuth2Session(client=LegacyApplicationClient(client_id=self.client_id),
                                  auto_refresh_url=self.authorization_url(instance),
                                  auto_refresh_kwargs=dict(client_id=self.client_id, client_secret=self.__secret))
            oauth.fetch_token(token_url=self.authorization_url(instance),
                              username=self.__username, password=self.__password, client_id=self.client_id,
                              client_secret=self.__secret, **kwargs)
            self.__password = None  # no longer need this.
            return oauth
        except ImportError:
            raise AuthenticationException('Install dependency requests-oauthlib')


class ServiceNowClientCredentialsFlow(ServiceNowFlow):
    """
    OAuth2 Client Credentials Grant Flow for ServiceNow.
    
    This flow is ideal for machine-to-machine authentication where no user context is needed.
    Only requires client_id and client_secret (no username/password).
    
    Example:
        >>> flow = ServiceNowClientCredentialsFlow('my_client_id', 'my_client_secret')
        >>> client = ServiceNowClient('dev12345', flow)
    """
    
    def __init__(self, client_id, client_secret):
        """
        Client Credentials flow authentication (OAuth 2.0)
        
        :param client_id: The OAuth application client ID
        :param client_secret: The OAuth application client secret
        """
        self.client_id = client_id
        self.__secret = client_secret
        self.__token = None
        self.__expires_at = None
        
    def authorization_url(self, authorization_base_url):
        """Generate the token endpoint URL"""
        return f"{authorization_base_url}/oauth_token.do"
    
    def _get_access_token(self, instance):
        """
        Request an access token from ServiceNow using client credentials.
        
        :param instance: The ServiceNow instance URL
        :return: Access token string
        :raises AuthenticationException: If token request fails
        """
        token_url = self.authorization_url(instance)
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
        }
        data = {
            'grant_type': 'client_credentials',
            'client_id': self.client_id,
            'client_secret': self.__secret
        }
        
        try:
            r = requests.post(token_url, headers=headers, data=data, timeout=30)
        except requests.exceptions.RequestException as e:
            raise AuthenticationException(f"Failed to connect to token endpoint: {e}")
        
        if r.status_code != 200:
            try:
                error_data = r.json()
                error_msg = error_data.get('error_description', error_data.get('error', r.text))
            except ValueError:
                error_msg = r.text
            raise AuthenticationException(
                f"Failed to obtain access token: {r.status_code} {r.reason} - {error_msg}"
            )
        
        try:
            token_data = r.json()
        except ValueError:
            raise AuthenticationException(f"Invalid JSON response from token endpoint: {r.text}")
        
        if 'access_token' not in token_data:
            raise AuthenticationException(f"No access_token in response: {token_data}")
        
        self.__token = token_data['access_token']
        # Use expires_in from response, default to 3600 seconds (1 hour) if not provided
        expires_in = token_data.get('expires_in', 3600)
        # Refresh 60 seconds before actual expiry to avoid edge cases
        self.__expires_at = int(time.time() + expires_in - 60)
        
        return self.__token
    
    def authenticate(self, instance: str, **kwargs) -> requests.Session:
        """
        Create and return an authenticated requests.Session with Bearer token.
        The session will automatically refresh the token when it expires.
        
        :param instance: The ServiceNow instance URL
        :param kwargs: Additional arguments (proxies, verify, etc.)
        :return: Authenticated requests.Session
        """
        session = requests.Session()
        
        # Get initial token
        if not self.__token or time.time() > (self.__expires_at or 0):
            self._get_access_token(instance)
        
        # Use a custom auth handler that refreshes tokens
        session.auth = _ClientCredentialsAuth(self, instance)
        
        return session


class _ClientCredentialsAuth(AuthBase):
    """
    Internal auth handler that automatically refreshes client credentials tokens.
    """
    
    def __init__(self, flow, instance):
        self._flow = flow
        self._instance = instance
    
    def __call__(self, request):
        # Check if token needs refresh
        if not self._flow._ServiceNowClientCredentialsFlow__token or \
           time.time() > (self._flow._ServiceNowClientCredentialsFlow__expires_at or 0):
            self._flow._get_access_token(self._instance)
        
        request.headers['Authorization'] = f"Bearer {self._flow._ServiceNowClientCredentialsFlow__token}"
        return request


class ServiceNowJWTAuth(AuthBase):

    def __init__(self, client_id, client_secret, jwt):
        """
        You must obtain a signed JWT from your OIDC provider, e.g. okta or auth0 or the like.
        We then use that JWT to issue an OAuth refresh token, that we then use to auth.
        """
        self.client_id = client_id
        self.__secret = client_secret
        self.__jwt = jwt
        self.__token = None
        self.__expires_at = None

    def _get_access_token(self, request):
        url = parse_url(request.url)
        token_url = f"{url.scheme}://{url.host}/oauth_token.do"
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
            'Authentication': f"Bearer {self.__jwt}"
        }
        data = {
            'grant_type': JWT_GRANT_TYPE,
            'client_id': self.client_id,
            'client_secret': self.__secret
        }
        r = requests.post(token_url, headers=headers, data=data)
        assert r.status_code == 200, f"Failed to auth, see syslogs {r.text}"
        data = r.json()
        expires = int(time.time()+data['expires_in'])
        return data['access_token'], expires

    def __call__(self, request):
        if not self.__token or time.time() > self.__expires_at:
            self.__token, self.__expires_at = self._get_access_token(request)
        request.headers['Authorization'] = f"Bearer {self.__token}"
        return request



"""
Asynchronous implementation of the pysnc package using httpx.AsyncClient.
"""

try:
    from .attachment import AsyncAttachment
    from .auth import (
        AsyncServiceNowFlow,
        AsyncServiceNowJWTAuth,
        AsyncServiceNowPasswordGrantFlow,
        AsyncServiceNowClientCredentialsFlow,
    )
    from .client import (
        AsyncAttachmentAPI,
        AsyncBatchAPI,
        AsyncServiceNowClient,
        AsyncTableAPI,
    )
    from .record import AsyncGlideRecord
except ImportError:
    raise ImportError("httpx is required for the asyncio module. Please install pysnc with the 'asyncio' extra.")

__all__ = [
    "AsyncServiceNowClient",
    "AsyncTableAPI",
    "AsyncBatchAPI",
    "AsyncAttachmentAPI",
    "AsyncGlideRecord",
    "AsyncAttachment",
    "AsyncServiceNowFlow",
    "AsyncServiceNowPasswordGrantFlow",
    "AsyncServiceNowClientCredentialsFlow",
    "AsyncServiceNowJWTAuth",
]
